# atomicswap

# Copyright 2022 Nicko van Someren
# SPDX: MIT
# See LICENSE.md for the full license text.

"""
atomicswap provides a simple API to swap two files on a filesystem atomically.
This means that after making the call to swap files at paths A and B, ether the
contents of A will be the former contents of B and the contents of B will be the
former contents of A, or an error will have occurred and the files will be as
they were before. The exchange is done in such a way that this there will not be
any point during the swapping when another process would be able to change the
outcome (although of course another process might be able to do so immediately
before or after).
"""

from sys import platform

if platform == "darwin":
    from .macos import swap
elif platform == "linux":
    from .linux import swap
else:
    raise NotImplementedError(f"Platform '{platform} is not currently supported")

# exchange is just an alias for swap
exchange = swap

__all__ = [exchange, swap]



# Tests for atomicswap

import os
from atomicswap import swap


A_DUMMY = "AAAA"
B_DUMMY = "BBBB"


def test_swap_str(tmp_path):
    file_a = str(tmp_path / "a")
    file_b = str(tmp_path / "b")

    with open(file_a, "w") as fh:
        fh.write(A_DUMMY)
    with open(file_b, "w") as fh:
        fh.write(B_DUMMY)

    swap(file_a, file_b)

    a_contents = open(file_a).read()
    assert(a_contents == B_DUMMY)
    b_contents = open(file_b).read()
    assert(b_contents == A_DUMMY)


def test_swap_path(tmp_path):
    file_a = tmp_path / "a"
    file_b = tmp_path / "b"

    with open(file_a, "w") as fh:
        fh.write(A_DUMMY)
    with open(file_b, "w") as fh:
        fh.write(B_DUMMY)

    swap(file_a, file_b)

    a_contents = open(file_a).read()
    assert(a_contents == B_DUMMY)
    b_contents = open(file_b).read()
    assert(b_contents == A_DUMMY)


def test_relative_cwd(tmp_path):
    os.chdir(tmp_path)

    file_a = "a"
    file_b = "b"

    with open(file_a, "w") as fh:
        fh.write(A_DUMMY)
    with open(file_b, "w") as fh:
        fh.write(B_DUMMY)

    swap(file_a, file_b)

    a_contents = open(file_a).read()
    assert(a_contents == B_DUMMY)
    b_contents = open(file_b).read()
    assert(b_contents == A_DUMMY)


def test_relative_dir_fd(tmp_path):
    dir_a = tmp_path / "A"
    dir_b = tmp_path / "B"

    os.mkdir(dir_a)
    os.mkdir(dir_b)

    with open(dir_a / "a", "w") as fh:
        fh.write(A_DUMMY)
    with open(dir_b / "b", "w") as fh:
        fh.write(B_DUMMY)

    dir_a_fd = os.open(dir_a, os.O_RDONLY)
    dir_b_fd = os.open(dir_b, os.O_RDONLY)

    swap("a", "b", first_dir_fd=dir_a_fd, second_dir_fd=dir_b_fd)

    a_contents = open(dir_a / "a").read()
    assert(a_contents == B_DUMMY)
    b_contents = open(dir_b / "b").read()
    assert(b_contents == A_DUMMY)


def test_relative_dir_mixed(tmp_path):
    os.chdir(tmp_path)

    dir_a = tmp_path / "A"
    dir_b = tmp_path / "B"

    os.mkdir(dir_a)
    os.mkdir(dir_b)

    with open(dir_a / "a", "w") as fh:
        fh.write(A_DUMMY)
    with open(dir_b / "b", "w") as fh:
        fh.write(B_DUMMY)

    dir_a_fd = os.open(dir_a, os.O_RDONLY)
    dir_b_fd = os.open(dir_b, os.O_RDONLY)

    swap("a", "B/b", first_dir_fd=dir_a_fd)

    a_contents = open(dir_a / "a").read()
    assert(a_contents == B_DUMMY)
    b_contents = open(dir_b / "b").read()
    assert(b_contents == A_DUMMY)

    swap("A/a", "b", second_dir_fd=dir_b_fd)

    a_contents = open(dir_a / "a").read()
    assert(a_contents == A_DUMMY)
    b_contents = open(dir_b / "b").read()
    assert(b_contents == B_DUMMY)

    swap("a", dir_b / "b", first_dir_fd=dir_a_fd)

    a_contents = open(dir_a / "a").read()
    assert(a_contents == B_DUMMY)
    b_contents = open(dir_b / "b").read()
    assert(b_contents == A_DUMMY)

    swap(dir_a / "a", "b", second_dir_fd=dir_b_fd)

    a_contents = open(dir_a / "a").read()
    assert(a_contents == A_DUMMY)
    b_contents = open(dir_b / "b").read()
    assert(b_contents == B_DUMMY)



# Implement atomic file swap/exchange for Linux.

# Copyright 2022 Nicko van Someren
# SPDX: MIT
# See LICENSE.md for the full license text.

from ctypes import CDLL, get_errno
from os import strerror

from typing import Union
from pathlib import PurePath

PathLike = Union[str, PurePath]

libc = CDLL("libc.so.6", use_errno=True)
syscall_function = libc.syscall
SYSCALL_RENAMEAT2 = 316
SWAP_FLAGS = 2
SWAP_AT_CWD = -100


def swap(
        first: PathLike, second: PathLike,
        *,
        first_dir_fd: int = SWAP_AT_CWD, second_dir_fd: int = SWAP_AT_CWD) -> None:
    """Atomically swap the files at the paths `first` and `second`. If either path is
    relative then these will be relative to the current working directory unless
    `first_dir_fd` and/or `second_dir_fd` are provided, in which case they must be
    file descriptors for directories from which the respective paths are relative.

    The function returns None, or raises an OSError is an error occurs.
    """
    result = syscall_function(
        SYSCALL_RENAMEAT2,
        first_dir_fd, str(first).encode(),
        second_dir_fd, str(second).encode(),
        SWAP_FLAGS
    )
    if result != 0:
        errno = get_errno()
        raise OSError(errno, strerror(errno))
